require "file_utils"
require "json"

class DODB::Tags(V) < DODB::Indexer(V)
	property name         : String
	property key_proc     : Proc(V, Array(String))
	getter   storage_root : String

	# Required to remove an entry in the DB.
	@storage : DODB::Storage(V)

	def initialize(@storage, @storage_root, @name, @key_proc)
		::Dir.mkdir_p indexing_directory
	end

	def check!(key, value, old_value)
		return true # Tags don’t have collisions or overloads.
	end

	def index(key, value)
		indices = key_proc.call(value)

		indices.each do |i|
			symlink = get_tagged_entry_path(i, key)
			Dir.mkdir_p ::File.dirname symlink
			# FIXME: Should not happen anymore. Should we remove this?
			::File.delete symlink if ::File.exists? symlink
			::File.symlink get_data_symlink(key), symlink
		end
	end

	def deindex(key, value)
		indices = key_proc.call(value)

		indices.each do |i|
			symlink = get_tagged_entry_path(i, key)
			::File.delete symlink
		end
	end

	def get_with_indice(tag : String) : Array(Tuple(V, Int32))
		r_value = Array(Tuple(V, Int32)).new

		tag_directory = indexing_directory tag

		return r_value unless Dir.exists? tag_directory

		Dir.each_child tag_directory do |child|
			r_value << {
				V.from_json(::File.read("#{tag_directory}/#{child}")),
				File.basename(child).gsub(/\.json$/, "").to_i
			}
		end

		r_value
	end

	def get_with_indices(keys : Array(String)) : Array(Tuple(V, Int32))
		r_value = Array(Tuple(V, Int32)).new
		keys.each do |tag|
			r_value.concat get_with_indice tag
		end
		r_value
	end

	def get(tag : String) : Array(V)
		get_with_indice(tag).map &.[0]
	end

	def get(keys : Array(String)) : Array(V)
		get_with_indices(keys.sort).map &.[0]
	end

	def delete(tag)
		delete tag, do true end
	end

	def delete(tag, &matcher)
		tag_directory = indexing_directory tag

		return unless Dir.exists? tag_directory

		Dir.each_child tag_directory do |child|
			path = "#{tag_directory}/#{child}"
			item = V.from_json ::File.read path

			if yield item
				key = get_key path
				@storage.delete key
			end
		end
	end

	private def get_key(path : String) : Int32
		::File.readlink(path)
			.sub(/\.json$/, "")
			.sub(/^.*\//,   "")
			.to_i
	end

	def indexing_directory : String
		"#{@storage_root}/tags/by_#{@name}"
	end

	private def indexing_directory(tag)
		"#{indexing_directory}/#{tag}"
	end

	private def get_tagged_entry_path(tag : String, key : String)
		"#{indexing_directory}/#{tag}/#{key}.json"
	end

	private def get_data_symlink(key : String)
		"../../../data/#{key}.json"
	end
end
